# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
"""
import numpy as np
import h5py
from exo_k.util.interp import rm_molec,unit_convert
from exo_k.settings import Settings
from exo_k.util.filenames import select_kwargs

class Hires_spectrum(object):
    """A class defining a Hires_spectrum object.
    """

    def __init__(self, filename, file_kdata_unit='unspecified', kdata_unit='unspecified',
        mult_factor=None, **kwargs):
        """Reads a high-resolution spectrum from a file (either hdf5 or ascii).

        Parameters
        ----------
            filename: str
                Full pathname to the file. Extension defines the format used.
            file_kdata_unit: str
                Specifies the unit for the opacity data in the file. This is needed
                for ascii formats as the units are not known.
                The type of quantity may differ whether we are handling cross sections (surface)
                or absorption coefficients (inverse length)
            kdata_unit: str
                Unit to convert to.
            mult_factor: float
                A multiplicative factor that can be applied to kdata (for example to correct for
                any dilution effect, or specific conversion).
        
        see :func:`read_ascii` for additional arguments to use with ascii files
        """
        
        self.filename=None
        self.kdata=None
        self.kdata_unit='unspecified'
        self.wns=None
        self.wn_unit='cm^-1'
        self.data_type=None # 'xsec' or 'abs_coeff'

        if filename.lower().endswith(('.hdf5', '.h5')):
            self.read_hdf5(filename)
        else:
            self.read_ascii(filename, **select_kwargs(kwargs,['skiprows','wn_column',
                'kdata_column','data_type']))

        if mult_factor is not None: self.kdata=self.kdata*mult_factor

        if self.data_type=='xsec':
            if (Settings()._convert_to_mks) and (kdata_unit is 'unspecified'):
                    kdata_unit='m^2/molecule'
        elif self.data_type=='abs_coeff':
            if (Settings()._convert_to_mks) and (kdata_unit is 'unspecified'):
                    kdata_unit='m^-1'
        else:
            raise RuntimeError("""Data type (xsec or abs_coeff) not recognized.
                You should specify it with the data_type keyword
                (and probably the file_kdata_unit as well).""")

        self.convert_kdata_unit(kdata_unit=kdata_unit,file_kdata_unit=file_kdata_unit)
       
    def read_ascii(self, filename, data_type=None, skiprows=0, wn_column=None,
        kdata_column=None):
        """Read native kspectrum format

        Parameters
        ----------
            filename: str
                Initial hires-spectrum filename.
            data_type: 'xsec' or 'abs_coeff'
                Whether the data read are cross-sections or absorption coefficients.
            skiprows: int, optional
                Number of header lines to skip. For the latest Kspectrum format,
                the header is skipped automatically.
            wn_column/kdata_column: int, optional
                Number of column to be read for wavenumber and kdata in python convention
                (0 is first, 1 is second, etc.)
        """
        if data_type is None:
            raise RuntimeError("You did not provide a data_type ('xsec' or 'abs_coeff')")
        self.data_type=data_type
        self.filename=filename
        with open(filename, 'r') as file:
            tmp = file.readline().split()
            if tmp[0]=='Pressure':
                #File treated as a Kspectrum like format, implying P in atm, Kdata in m^-1
                for _ in range(4):
                    file.readline()
                nb_mol = int(file.readline().split()[0])
                skiprows=skiprows+9+5*nb_mol
                if kdata_column is None:
                    kdata_column=2
            else:
                if kdata_column is None: kdata_column=1

        if wn_column is None: wn_column=0
        raw=np.genfromtxt(filename, skip_header=skiprows,
            usecols=(wn_column,kdata_column), names=('wns','kdata')) 
        self.kdata=raw['kdata']
        self.wns=raw['wns']


    def write_hdf5(self, filename):
        """Writes kspectrum file to hdf5
        """
        if not filename.lower().endswith(('.hdf5', '.h5')):
            filename=filename+'.h5'
        f = h5py.File(filename, 'w')
        f.attrs["data_type"] = self.data_type
        f.create_dataset("wns", data=self.wns,compression="gzip")
        f["wns"].attrs["units"]=self.wn_unit
        f.create_dataset("kdata", data=self.kdata,compression="gzip")
        f["kdata"].attrs["units"]=self.kdata_unit
        f.close()    

    def read_hdf5(self, filename):
        """Reads kspectrum file from hdf5
        """
        self.filename=filename
        f = h5py.File(filename, 'r')
        self.data_type=f.attrs['data_type']
        self.wns=f['wns'][...]
        if 'units' in f['wns'].attrs.keys():
            self.wn_unit=f['wns'].attrs['units']
        self.kdata=f['kdata'][...]
        self.kdata_unit=f['kdata'].attrs['units']
        f.close()

    def convert_kdata_unit(self, kdata_unit='unspecified', file_kdata_unit='unspecified'):
        """Converts kdata to a new unit (in place)

        Parameters
        ----------
            kdata_unit: str
                String to identify the units to convert to. Accepts 'cm^2', 'm^2'
                or any surface unit recognized by the astropy.units library.
                If ='unspecified', no conversion is done.
                In general, kdata should be kept in 'per number' or 'per volume'
                units (as opposed to 'per mass' units) as composition will
                always be assumed to be a number or volume mixing ratio.
                Opacities per unit mass are not supported yet.
                Note that you do not need to specify the '/molec' or
                '/molecule' in the unit.
            file_kdata_unit : str
                String to specify the current kdata unit if it is unspecified or if 
                you have reasons to believe it is wrong (e.g. you just read a file where
                you know that the kdata grid and the kdata unit do not correspond)
        """
        #if kdata_unit==file_kdata_unit and self.kdata_unit != 'unspecified': return
        tmp_k_u_in=rm_molec(file_kdata_unit)
        tmp_k_u_out=rm_molec(kdata_unit)
        tmp_k_u_file=rm_molec(self.kdata_unit)
        self.kdata_unit, conversion_factor=unit_convert(  \
            'kdata_unit',unit_file=tmp_k_u_file,unit_in=tmp_k_u_in,unit_out=tmp_k_u_out)
        if self.data_type=='xsec':
            self.kdata_unit=self.kdata_unit+'/molecule'
        else:
            self.kdata_unit=self.kdata_unit
        self.kdata=self.kdata*conversion_factor

    def __repr__(self):
        """Method to output header
        """
        output="""
        file         : {file}
        data type    : {dtype}
        data unit    : {ku}
        size         : {size}
        wns          : {wns}
        wn units     : {wnu}
        """.format(file=self.filename, ku=self.kdata_unit,
        wns=self.wns, wnu=self.wn_unit, dtype=self.data_type, size= self.wns.size)
        return output

    def plot_spectrum(self, ax, x_axis='wls',
            xscale=None, yscale=None, **kwarg):
        """Plot the spectrum
        
        Parameters
        ----------
            ax : :class:`pyplot.Axes`
                A pyplot axes instance where to put the plot.
            x_axis: str, optional
                If 'wls', x axis is wavelength. Wavenumber otherwise.
            x/yscale: str, optional
                If 'log' log axes are used.
        """
        toplot=self.kdata
        if x_axis == 'wls':
            ax.plot(self.wls,toplot,**kwarg)
            ax.set_xlabel('Wavelength (micron)')
        else:
            ax.plot(self.wns,toplot,**kwarg)
            ax.set_xlabel('Wavenumber (cm$^{-1}$)')
        if self.data_type=='xsec':
            ax.set_ylabel('Cross section ('+self.kdata_unit+')')
        else:
            ax.set_ylabel('Abs. Coefficient ('+self.kdata_unit+')')
        ax.grid(True)
        if xscale is not None: ax.set_xscale(xscale)
        if yscale is not None: ax.set_yscale(yscale)

    @property
    def wls(self):
        """Returns the wavelength array for the bin centers
        """
        return 10000./self.wns



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

Library of useful functions for handling filenames
"""
import array
import numpy as np
import h5py

class EndOfFile(Exception):
    """Error for an end of file
    """

def select_kwargs(kwargs, filter_keys_list=[]):
    """Function to select only some keyword arguments from a
    kwargs dict to pass to a function afterward.
    
    Parameters
    ----------
        kwargs: dict
            Dictionary of keyword arguments and values.
        filter_keys_list: list of str
            Names of the keys to select.

    Returns
    -------
        filtered_kwargs: dict
            A dictionary with only the selected keys (if they were
            present in kwargs).
        
    Examples
    --------
        >>> def func(allo=None):
        >>>     print(allo)
        >>>
        >>> def bad_global_func(**kwargs):
        >>>     print(kwargs)
        >>>     func(**kwargs)
        >>>
        >>> def good_global_func(**kwargs):
        >>>     print(kwargs)
        >>>     func(**select_kwargs(kwargs,['allo']))

        >>> bad_global_func(allo=3.,yeah=None)        
        {'allo': 3.0, 'yeah': None}
        TypeError: func() got an unexpected keyword argument 'yeah'

        >>> good_global_func(allo=3.,yeah=None)        
        {'allo': 3.0, 'yeah': None}
        3.0

    """
    filtered_kwargs=dict()
    for key in filter_keys_list:
        if key in kwargs.keys():
            filtered_kwargs[key]=kwargs[key]
    return filtered_kwargs

def create_fname_grid_Kspectrum_LMDZ(Np, Nt, Nx=None, suffix='', nb_digit=3):
    """Creates a grid of filenames consistent with Kspectrum/LMDZ
    from the number of pressure, temperatures (, and vmr) points (respectively) in the grid.

    Parameters
    ----------
        Np, Nt: int
            Number of Pressure and temperature points.
        Nx: int, optional
            Number of vmr points if there is a variable gas
        suffix: str, optional
            String to add behind the 'k001'
        nb_digit: int, optional
            Total number of digit including the zeros ('k001' is 3)
    
    Returns
    -------
        list of str
            List of the files in the right order and formating to be 
            given to :func:`exo_k.ktable.Ktable.hires_to_ktable` or 
            :func:`exo_k.xtable.Xtable.hires_to_xtable`.

    Examples
    --------

        >>> exo_k.create_fname_grid_Kspectrum_LMDZ(2,3)                  
        array([['k001', 'k002', 'k003'],
        ['k004', 'k005', 'k006']], dtype='<U4')

        >>> exo_k.create_fname_grid_Kspectrum_LMDZ(2,3,suffix='.h5')                  
        array([['k001.h5', 'k002.h5', 'k003.h5'],
        ['k004.h5', 'k005.h5', 'k006.h5']], dtype='<U7')

    """    
    res=[]
    ii=1
    if Nx is None:
        for _ in range(Np):
            for _ in range(Nt):
                res.append('k'+str(ii).zfill(nb_digit)+suffix)
                ii+=1
        return np.array(res).reshape((Np,Nt))
    else:
        for _ in range(Nx):
            for _ in range(Np):
                for _ in range(Nt):
                    res.append('k'+str(ii).zfill(nb_digit)+suffix)
                    ii+=1
        res=np.array(res).reshape((Nx,Np,Nt))
        return np.transpose(res,(2,1,0))

def read_nemesis_binary(filename):
    """reads a nemesis binary file.
    """
    f = open(filename, mode='rb')
    int_array = array.array('i')
    float_array = array.array('f')
    int_array.fromfile(f, 2)
    irec0, Nw=int_array[-2:]
    float_array.fromfile(f, 3)
    wl_min, dwl, FWHM = float_array[-3:]
    int_array.fromfile(f, 5)
    Np, Nt, Ng = int_array[-5:-2]
    float_array.fromfile(f, Ng)
    ggrid = np.array(float_array[-Ng:])
    float_array.fromfile(f, Ng)
    weights = np.array(float_array[-Ng:])
    float_array.fromfile(f, 2)
    float_array.fromfile(f, Np)
    pgrid = np.array(float_array[-Np:])
    float_array.fromfile(f, Nt)
    tgrid = np.array(float_array[-Nt:])
    ntot=Nw*Np*Nt*Ng
    if dwl>=0.: #regular grid
        wls=wl_min+np.arange(Nw)*dwl
    else:
        float_array.fromfile(f, Nw)
        wls = np.array(float_array[-Nw:])
    wns=10000./wls[::-1]
    f.close()
    f = open(filename, mode='rb') # restart to start reading kdata at record number irec0
    kdata=array.array('f')
    kdata.fromfile(f, irec0-1+ntot)
    kdata=np.reshape(kdata[-ntot:],(Nw,Np,Nt,Ng))[::-1]*1.e-20
    kdata=kdata.transpose(1,2,0,3)
    f.close()
    return Np,Nt,Nw,Ng,pgrid,tgrid,wns,ggrid,weights,kdata

def convert_exo_transmit_to_hdf5(file_in, file_out, mol='unspecified'):
    """Converts exo_transmit like spectra to hdf5 format for speed and space.

    Parameters
    ----------
        file_in: str
            Initial exo_transmit filename.
        file_out: str
            Name of the final hdf5 file to be created.
    """
    tmp_wlgrid=[]
    tmp_kdata=[]
    with open(file_in, 'r') as file:
        tmp = file.readline().split()
        tgrid=np.array([float(ii) for ii in tmp])
        tmp = file.readline().split()
        pgrid=np.array([float(ii) for ii in tmp])
        Np=pgrid.size
        Nt=tgrid.size
        while True:
            line=file.readline()
            if line is None or line=='': break
            tmp_wlgrid.append(float(line.split()[0])) # wavelength in m
            tmp_kdata.append([])
            for _ in range(Np):
                tmp_kdata[-1].append(file.readline().split()[1:])
    tmp_wlgrid=np.array(tmp_wlgrid)
    Nw=tmp_wlgrid.size
    kdata=np.zeros((Np,Nt,Nw))
    for iP in range(Np):
        for iT in range(Nt):
            for iW in range(Nw):
                kdata[iP,iT,iW]=tmp_kdata[Nw-iW-1][iP][iT]
    print(kdata.shape)
    wns=0.01/tmp_wlgrid[::-1]
    
    if not file_out.lower().endswith(('.hdf5', '.h5')):
        fullfilename=file_out+'.hdf5'
    else:
        fullfilename=file_out
    compression="gzip"
    f = h5py.File(fullfilename, 'w')
    f.attrs["mol_name"] = mol
    f.create_dataset("p", data=pgrid,compression=compression)
    f["p"].attrs["units"] = 'Pa'
    f.create_dataset("t", data=tgrid,compression=compression)
    f.create_dataset("xsecarr", data=kdata,compression=compression)
    f["xsecarr"].attrs["units"] = 'm^2'
    f.create_dataset("bin_edges", data=wns,compression=compression)
    f.close()    

####### OLD FUNCTIONS ###################
def convert_old_kspectrum_to_hdf5(file_in, file_out, skiprows=0):
    """Deprecated.
    
    Converts kspectrum like spectra to hdf5 format for speed and space

    Parameters
    ----------
        file_in: str
            Initial kspectrum filename.
        file_out: str
            Name of the final hdf5 file to be created.
    """
    wn_hr,k_hr=np.loadtxt(file_in,skiprows=skiprows,unpack=True) 
    f = h5py.File(file_out, 'w')
    f.create_dataset("wns", data=wn_hr,compression="gzip")
    f.create_dataset("k", data=k_hr,compression="gzip")
    f.close()   



"""
@author: jeremy leconte

A class with some basic functions for all objects with a spectral dimension
"""

class Spectral_object(object):
    """A class with some basic functions for all objects with a spectral dimension
    """

    @property
    def wls(self):
        """Returns the wavelength array for the bin centers
        """
        return 10000./self.wns



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
"""
import numpy as np
from .ktable import Ktable
from .xtable import Xtable
from .gas_mix import Gas_mix
from .settings import Settings
from .util.interp import rm_molec

class Kdatabase(object):
    """This object contains mainly a dictionary of individual :class:`~exo_k.ktable.Ktable`
    or :class:`~exo_k.xtable.Xtable` objects for each molecule. 

    In addition, the informations about the P, T, Wn, g grids
    are reloaded as atributes of the Kdatabase object.
    """

    def __init__(self, molecules, *str_filters, search_path=None,
        remove_zeros=True, **kwargs):
        """Initializes k coeff tables and supporting data from a list of molecules

        Parameters
        ----------
            molecules: list or dict
                * If a list of molecules is provided,
                  the file starting with the molecule name and containing all the str_filters
                  are searched in the global search path defined in
                  :class:`~exo_k.settings.Settings`.
                * If a dictionary is provided, the keys are the molecules to load,
                  and the values are the path to the corresponding file.
                  If a None value is given, the str_filters will be used as above.
            search_path: str, optional
                If search_path is provided, it locally overrides the global _search_path settings
                and only files in search_path are returned.            

        See also the options of :class:`~exo_k.ktable.Ktable`

        >>> Kdatabase(None)

        Loads an empty database to be filled later.
        """
        self.ktables={}
        self._settings=Settings()
        self.consolidated_wn_grid=True
        self.consolidated_PT_grid=True
        self.consolidated_p_unit=True
        self.consolidated_kdata_unit=True

        self.molecules=None
        if molecules is None:
            return
        if isinstance(molecules,list):
            for mol in molecules:
                try:
                    tmp_ktable=Ktable(*str_filters, mol=mol,
                        remove_zeros=remove_zeros, search_path=search_path, **kwargs)
                except:
                    tmp_ktable=Xtable(*str_filters, mol=mol,
                        remove_zeros=remove_zeros, search_path=search_path, **kwargs)
                self.add_ktables(tmp_ktable)
        else: #then we assume a dict
            for mol,filename in molecules.items():
                try:
                    # below, we still provide  *([mol+delim]+list(str_filters)) 
                    # as an input in case filename is None
                    tmp_ktable=Ktable(*str_filters, filename=filename, mol=mol,
                        remove_zeros=remove_zeros, search_path=search_path, **kwargs)
                except:
                    tmp_ktable=Xtable(*str_filters, filename=filename, mol=mol,
                        remove_zeros=remove_zeros, search_path=search_path, **kwargs)
                self.add_ktables(tmp_ktable)
        
    def add_ktables(self, *ktables):
        """Adds as many :class:`~exo_k.ktable.Ktable` or :class:`~exo_k.xtable.Xtable`
        to the database as you want.

        Parameters
        ----------
            ktables: :class:`Ktable` or :class:`Xtable` objects
                Tables to be added. 
        """
        for tmp_ktable in ktables:
            if self.molecules is None:
                self.ktables[tmp_ktable.mol]=tmp_ktable
                self.kdata_unit=tmp_ktable.kdata_unit
                self.pgrid=tmp_ktable.pgrid
                self.p_unit=tmp_ktable.p_unit
                self.logpgrid=tmp_ktable.logpgrid
                self.tgrid=tmp_ktable.tgrid
                self.wns=tmp_ktable.wns
                self.wnedges=tmp_ktable.wnedges
                self.Np=tmp_ktable.Np
                self.Nt=tmp_ktable.Nt
                self.Nw=tmp_ktable.Nw
                self.Ng=tmp_ktable.Ng
                if tmp_ktable.Ng is not None:
                    self.Ng=tmp_ktable.Ng
                    self.weights=tmp_ktable.weights
                    self.ggrid=tmp_ktable.ggrid
                    self.gedges=tmp_ktable.gedges
            else:
                if (self.Ng is None)^(tmp_ktable.Ng is None): raise RuntimeError( \
                    'All elements in a database must have the same type (Ktable or Xtable).')
                if (self.Ng is not None) and (not np.array_equal(tmp_ktable.ggrid,self.ggrid)):
                    raise RuntimeError('All Ktables in a database must have the same g grid.')
                if self.p_unit != tmp_ktable.p_unit:
                    print("""Carefull, not all tables have the same p unit.
                        You'll need to use convert_p_unit""")
                    self.consolidated_p_unit=False
                if rm_molec(self.kdata_unit) != rm_molec(tmp_ktable.kdata_unit):
                    print("""Carefull, not all tables have the same kdata unit.
                        You'll need to use convert_kdata_unit""")
                    self.consolidated_kdata_unit=False
                self.ktables[tmp_ktable.mol]=tmp_ktable
                if not np.array_equal(tmp_ktable.wns,self.wns):
                    self.consolidated_wn_grid=False
                    print("""Carefull, not all tables have the same wevelength grid.
                        You'll need to use bin_down""")
                    self.wns    = None
                    self.wnedges= None
                    self.Nw     = None
                if not (np.array_equal(tmp_ktable.pgrid,self.pgrid) \
                    and np.array_equal(tmp_ktable.tgrid,self.tgrid)) :
                    self.consolidated_PT_grid=False
                    self.pgrid   = None
                    self.logpgrid= None
                    self.tgrid   = None
                    self.Np      = None
                    self.Nt      = None
                    print("""Carefull, not all tables have the same PT grid.
                        You'll need to use remap_logPT""")
            self.molecules=list(self.ktables.keys())

    def copy(self):
        """Creates a new instance of :class:`~exo_k.kdatabase.Kdatabase`
        object and (deep) copies data into it
        """
        res=Kdatabase(None)
        if self.molecules is not None:
            for ktab in self.ktables.values():
                res.add_ktables(ktab.copy())
        return res

    def __repr__(self):
        """Method to output
        """
        output='The available molecules are: \n'
        for mol, ktab in self.ktables.items():
            output+=mol+'->'+ktab.filename+'\n'
        if self.consolidated_wn_grid:
            output+='All tables share a common spectral grid\n'
        else:
            output+='All tables do NOT have common spectral grid\n'
            output+='You will need to run bin_down or sample before using the database\n'
        if self.consolidated_PT_grid:
            output+='All tables share a common logP-T grid\n'
        else:
            output+='All tables do NOT have common logP-T grid\n'
            output+='You will need to run remap_logPT to perform some operations\n'
        if self.consolidated_p_unit:
            output+='All tables share a common p unit\n'
        else:
            output+='All tables do NOT have common p unit\n'
            output+='You will need to run convert_p_unit to perform some operations\n'
        if self.consolidated_kdata_unit:
            output+='All tables share a common kdata unit\n'
        else:
            output+='All tables do NOT have common kdata unit\n'
            output+='You will need to run convert_kdata_unit to perform some operations\n'

        return output

    def __getitem__(self, molecule):
        """Overrides getitem so as to access directly a Ktable with Kdatabase['mol']
        """
        if molecule not in self.ktables.keys():
            raise KeyError('The requested molecule is not available.')
        return self.ktables[molecule]

    @property
    def wls(self):
        """Returns the wavelength array for the bin centers
        """
        return 10000./self.wns

    @property
    def wledges(self):
        """Returns the wavelength array for the bin edges
        """
        return 10000./self.wnedges

    def remap_logPT(self, logp_array=None, t_array=None):
        """Applies the bin_down method to all the tables in the database. This can be used 
        to put all the tables onthe same PT grid.

        See :func:`exo_k.data_table.Data_table.remap_logPT` for details.
        """
        if not self.consolidated_p_unit: raise RuntimeError( \
            """All tables in the database should have the same p unit to proceed.
            You should probably use convert_p_unit().""")
        for mol in self.molecules:
            self.ktables[mol].remap_logPT(logp_array=logp_array,t_array=t_array)
        self.logpgrid=np.array(logp_array, dtype=float)
        self.pgrid   =10**self.logpgrid
        self.tgrid   =np.array(t_array)
        self.Np      =self.logpgrid.size
        self.Nt      =self.tgrid.size
        self.consolidated_PT_grid=True

    def bin_down(self, wnedges=None, **kwargs):
        """Applies the bin_down method to all the tables in the database. This can be used 
        to put all the tables onthe same wavenumber grid.

        See :func:`exo_k.ktable.Ktable.bin_down` or :func:`exo_k.xtable.Xtable.bin_down`
        for details.
        """
        first=True
        for mol in self.molecules:
            self.ktables[mol].bin_down(wnedges=wnedges, **kwargs)
            if first:
                self.wns=self.ktables[mol].wns
                self.wnedges=self.ktables[mol].wnedges
                self.Nw=self.ktables[mol].Nw
                if self.Ng is not None:
                    self.ggrid=self.ktables[mol].ggrid
                    self.weights=self.ktables[mol].weights
                    self.Ng=self.ktables[mol].Ng
                self.consolidated_wn_grid=True

    def bin_down_cp(self, wnedges=None, **kwargs):
        """Creates a copy of the database and bins it down.

        Parameters
        ----------
            See `bin_down` for details on parameters

        Returns
        -------
            :class:`Kdatabase` object
                The binned down database
        """
        res=self.copy()
        res.bin_down(wnedges=wnedges, **kwargs)
        return res

    def sample(self, wngrid, **kwargs):
        """Applies the bin_down method to all the tables in the database. This can be used 
        to put all the tables onthe same wavenumber grid.

        See Ktable.bin_down() or Xtable.bin_down() for details.
        """
        first=True
        if self.Ng is not None: raise RuntimeError('sample is only available for Xtable objects.')
        for mol in self.molecules:
            self.ktables[mol].sample(wngrid,**kwargs)
            if first:
                self.wns=self.ktables[mol].wns
                self.wnedges=self.ktables[mol].wnedges
                self.Nw=self.ktables[mol].Nw
                self.consolidated_wn_grid=True

    def sample_cp(self, wngrid, **kwargs):
        """Creates a copy of the database and re-samples it.

        Parameters
        ----------
            See `sample` for details on parameters

        Returns
        -------
            :class:`Kdatabase` object
                The re-sampled database
        """
        if self.Ng is not None: raise RuntimeError('sample is only available for Xtable objects.')
        res=self.copy()
        res.sample(wngrid, **kwargs)
        return res

    def clip_spectral_range(self, wn_range=None, wl_range=None):
        """Limits the data to the provided spectral range:

           * Wavenumber in cm^-1 if using wn_range argument
           * Wavelength in micron if using wl_range
        """
        first=True
        for mol in self.molecules:
            self.ktables[mol].clip_spectral_range(wn_range=wn_range, wl_range=wl_range)
            if first:
                self.wns=self.ktables[mol].wns
                self.wnedges=self.ktables[mol].wnedges
                self.Nw=self.ktables[mol].Nw

    def convert_p_unit(self, p_unit='unspecified'):
        """Converts pressure unit of all Data_tables.
        See Data_table.convert_p_unit() for details.
        """
        first=True
        for mol in self.molecules:
            self[mol].convert_p_unit(p_unit=p_unit)
            if first:
                self.p_unit=self[mol].p_unit
        self.consolidated_p_unit=True

    def convert_kdata_unit(self, kdata_unit='unspecified'):
        """Converts kdata unit of all Data_tables.
        See Data_table.convert_kdata_unit() for details.
        """
        first=True
        for mol in self.molecules:
            self[mol].convert_kdata_unit(kdata_unit=kdata_unit)
            if first:
                self.kdata_unit=self[mol].kdata_unit
        self.consolidated_kdata_unit=True

    def convert_to_mks(self):
        """Converts units of all `Data_table`s to MKS
        """
        first=True
        for mol in self.molecules:
            self[mol].convert_to_mks()
            if first:
                self.p_unit=self[mol].p_unit
                self.kdata_unit=self[mol].kdata_unit
        self.consolidated_p_unit=True
        self.consolidated_kdata_unit=True

    def create_mix_ktable(self, composition, inactive_species=[]):
        """creates the kdata table for a mix of molecules.

        Parameters
        ----------
            composition: dict
                Keys are the molecule names (they must match the names in the database).
                Values are either numbers or arrays of volume mixing ratios
                with shape (pgrid.size,tgrid.size).
                This composition will instantiate a :class:`Gas_mix` object.
                In particular, if a value is 'background', this gas will
                be used to fill up to sum(vmr)=1 (See chemistry.gas_type for details).
                For each (P,T) point, the sum of all the mixing ratios
                should be lower or equal to 1.
                If it is lower, it is assumed that the rest of the gas is transparent.
            inactive_species: list, optional
                List the gases that are in composition but for which we do not want the 
                opacity to be accounted for. 

        Returns
        -------
            Ktable object
                A new ktable for the mix
        """
        if self.Ng is None: raise RuntimeError( \
            """Create_mix_ktable cannot work with a database of cross sections.
            Please load correlated-k data.""")
        if not self.consolidated_wn_grid: raise RuntimeError( \
            """All tables in the database should have the same wavenumber grid to proceed.
            You should probably use bin_down().""")
        if not self.consolidated_PT_grid: raise RuntimeError( \
            """All tables in the database should have the same PT grid to proceed.
            You should probably use remap_logPT().""")
        if not self.consolidated_p_unit: raise RuntimeError( \
            """All tables in the database should have the same p unit to proceed.
            You should probably use convert_p_unit().""")
        if not self.consolidated_kdata_unit: raise RuntimeError( \
            """All tables in the database should have the same p unit to proceed.
            You should probably use convert_kdata_unit().""")
        mol_to_be_done=set(composition.keys())
        mol_to_be_done=mol_to_be_done-set(inactive_species)
        if not mol_to_be_done:
            print("""You are creating a mix without any active gas:
                This will be awfully transparent""")
            res=self[self.molecules[0]].copy(cp_kdata=False)
            res.kdata=np.zeros(res.shape)
            return res
        gas_mixture=Gas_mix(composition)
        if all(elem in self.molecules for elem in mol_to_be_done):
            print('I have all the requested molecules in my database')
            print(mol_to_be_done)
        else:
            print('Do not have all the molecules in my database')
            mol_to_be_done=mol_to_be_done.intersection(set(self.molecules))
            print('Molecules to be treated: ', mol_to_be_done)
        first_mol=True
        for mol in mol_to_be_done:
            if first_mol:
                res=self.ktables[mol].copy(cp_kdata=True)
                try:
                    res.kdata=res.vmr_normalize(gas_mixture[mol])
                except TypeError:
                    print('gave bad mixing ratio format to vmr_normalize')
                    raise TypeError('bad mixing ratio type')            
                if len(mol_to_be_done)==1:
                    print('only 1 molecule:',mol)
                    return res
                first_mol=False
            else:
              print('treating molecule ',mol)
              res.kdata=res.RandOverlap(self.ktables[mol], None, gas_mixture[mol])
              # no need to re normalize with respect to 
              # the abundances of the molecules already done.
        return res  

    def create_mix_ktable5d(self, bg_comp={}, vgas_comp={}, x_array=None,
            bg_inac_species=[], vgas_inac_species=[], **kwargs):
        """Creates a Ktable5d for a mix of molecules with a variable gas.
        In essence, the regular create_mix_ktable is called to create
        two mixes:
        * The background mix specified by composition=bg_comp, inactive_species=bg_inac_species
        * The variable gas specified by composition=vgas_comp, inactive_species=vgas_inac_species

        See create_mix_ktable for details.

        These two gases are then mixed together for an array of vmr x_array where
        var_gas has a vmr of x_array and the background gas has a vmr of 1.-x_array

        Returns
        -------
            Ktable5d object
                A new ktable for the mix
        """
        if x_array is None:
            raise RuntimeError('x_array is None: pas bien!!!')
        background_mix=self.create_mix_ktable(bg_comp, inactive_species=bg_inac_species)
        var_gas_mix=self.create_mix_ktable(vgas_comp, inactive_species=vgas_inac_species)
        ktab5d=var_gas_mix.copy(ktab5d=True)
        ktab5d.xgrid=np.array(x_array)
        ktab5d.Nx=ktab5d.xgrid.size
        print('shape of the output Ktable5d (p,t,x,wn,g):', ktab5d.shape)
        new_kdata=np.zeros(ktab5d.shape)
        for iX, vmr in enumerate(ktab5d.xgrid):
            new_kdata[:,:,iX,:,:]=var_gas_mix.RandOverlap(background_mix, vmr, 1.-vmr, **kwargs)
        ktab5d.set_kdata(new_kdata)
        return ktab5d



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
"""
import os.path
from math import log10
import pickle
import h5py
import numpy as np
import astropy.units as u
from .data_table import Data_table
from .util.interp import rebin_ind_weights
from .util.cst import KBOLTZ
from .hires_spectrum import Hires_spectrum
from .util.filenames import create_fname_grid_Kspectrum_LMDZ, select_kwargs

class Xtable(Data_table):
    """A class that handles tables of cross sections.
    """

    def __init__(self, *filename_filters, filename=None,
        p_unit='unspecified', file_p_unit='unspecified',
        kdata_unit='unspecified', file_kdata_unit='unspecified',
        remove_zeros=False, search_path=None, mol=None):
        """Initializes cross section table and
        supporting data from a file based on its extension.

        Parameters
        ----------
            filename: str
                Relative or absolute path to the input file.
            filename_filters: sequence of string
                As many strings as necessary to uniquely define
                a file in the global search path defined in
                :class:`~exo_k.settings.Settings`.
                This path will be searched for a file
                with all the filename_filters in the name.
                The filename_filters can contain '*'.

        If there is no filename or filename_filters provided,
        just creates an empty object to be filled later

        See :class:`~exo_k.ktable.Ktable` __init__ mehthod for documentation on
        `p_unit`, `file_p_unit`, `kdata_unit`, `file_kdata_unit`, `remove_zeros`,
        `search_path`, and `mol` keywords.
        """
        super().__init__()
        if filename is not None:
            self.filename=filename
        elif filename_filters or mol is not None:
            self.filename=self._settings.list_files(*filename_filters, molecule = mol,
                                 only_one=True, search_path=search_path)[0]
        if self.filename is not None:
            if self.filename.lower().endswith('pickle'):
                self.read_pickle(filename=self.filename, mol=mol)
            elif self.filename.lower().endswith(('.hdf5', '.h5')):
                self.read_hdf5(filename=self.filename, mol=mol)
            elif self.filename.lower().endswith(('.dat')):
                self.read_exo_transmit(filename=self.filename, mol=mol)
            else:
                raise NotImplementedError("""Requested format not recognized.
                Currently recognized formats are Exomol .pickle, .hdf5, and Exo_Transmit .dat.""")

        super().finalize_init(p_unit=p_unit, file_p_unit=file_p_unit,
            kdata_unit=kdata_unit, file_kdata_unit=file_kdata_unit,
            remove_zeros=remove_zeros)

    @property
    def shape(self):
        """Returns the shape of self.kdata
        """
        return np.array([self.Np,self.Nt,self.Nw])

    def read_pickle(self, filename=None, mol=None):
        """Initializes xsec table and supporting data from an Exomol pickle file

        Parameters
        ----------
            filename : str
                Relative or absolute name of the input pickle file
            mol : str, optional
                Overrides the name of the molecule to be put in the :class:`Xtable` object.
        """
        if filename is None: raise TypeError("You should provide an input pickle filename")
        pickle_file=open(filename,'rb')
        raw=pickle.load(pickle_file, encoding='latin1')
        pickle_file.close()
        
        if mol is not None: 
            self.mol=mol
        else:
            self.mol=raw['name']
            if self.mol=='H2OP': self.mol='H2O'

        self.tgrid=raw['t']
        self.wns=raw['wno']
        self.wnedges=np.concatenate( \
            ([self.wns[0]],(self.wns[:-1]+self.wns[1:])*0.5,[self.wns[-1]]))
        self.logk=False

        #deals with the p grid and units
        if 'p_unit' in raw.keys():
            self.p_unit=raw['p_unit']
        else:
            self.p_unit='bar'
        self.pgrid=raw['p']
        self.logpgrid=np.log10(self.pgrid)

        #deals with the k grid and units
        if 'kdata_unit' in raw.keys():
            self.kdata_unit=raw['kdata_unit']
        else:
            self.kdata_unit='cm^2/molec'
        self.kdata=raw['xsecarr']

        if 'wn_unit' in raw.keys(): self.wn_unit=raw['wn_unit']

        self.Np,self.Nt,self.Nw=self.kdata.shape

    def write_pickle(self, filename):
        """Saves data in a pickle format

        Parameters
        ----------
            filename: str
                Relative or absolute name of the file to be created and saved
        """
        fullfilename=filename
        if not filename.lower().endswith('.pickle'): fullfilename=filename+'.pickle'
        pickle_file=open(fullfilename,'wb')
        dictout={'name':self.mol,
                 'p':self.pgrid,
                 'p_unit':self.p_unit,
                 't':self.tgrid,
                 'wno':self.wns,
                 'wn_unit':self.wn_unit,
                 'xsecarr':self.kdata,
                 'kdata_unit':self.kdata_unit}
        #print(dictout)
        pickle.dump(dictout,pickle_file,protocol=-1)
        pickle_file.close()

    def read_hdf5(self, filename=None, mol=None):
        """Initializes k coeff table and supporting data from an Exomol hdf5 file

        Parameters
        ----------
            filename : str
                Name of the input hdf5 file
            mol : str, optional
                Overrides the name of the molecule to be put in the :class:`Xtable` object.
        """
        if (filename is None or not filename.lower().endswith(('.hdf5', '.h5'))):
            raise RuntimeError("You should provide an input hdf5 file")
        f = h5py.File(filename, 'r')
        if 'mol_name' in f:
            self.mol=f['mol_name'][()]
        elif 'mol_name' in f.attrs:
            self.mol=f.attrs['mol_name']
        else:
            self.mol=os.path.basename(filename).split(self._settings._delimiter)[0]
        if isinstance(self.mol, np.ndarray): self.mol=self.mol[0]
        if mol is not None:
            self.mol=mol
        if 'bin_edges' in f:
            self.wns=f['bin_edges'][...]
            if 'units' in f['bin_edges'].attrs:
                self.wn_unit=f['bin_edges'].attrs['units']
        else:
            self.wns=f['bin_centers'][...]
            if 'units' in f['bin_centers'].attrs:
                self.wn_unit=f['bin_centers'].attrs['units']
        self.wnedges=np.concatenate(  \
            ([self.wns[0]],(self.wns[:-1]+self.wns[1:])*0.5,[self.wns[-1]]))
        self.kdata=f['xsecarr'][...]
        self.kdata_unit=f['xsecarr'].attrs['units']
        self.tgrid=f['t'][...]
        self.pgrid=f['p'][...]
        self.logpgrid=np.log10(self.pgrid)
        self.p_unit=f['p'].attrs['units']
        self.logk=False
        f.close()  
        self.Np,self.Nt,self.Nw=self.kdata.shape

    def write_hdf5(self, filename):
        """Saves data in a hdf5 format

        Parameters
        ----------
            filename: str
                Name of the file to be created and saved
        """
        fullfilename=filename
        if not filename.lower().endswith(('.hdf5', '.h5')):
            fullfilename=filename+'.hdf5'
        compression="gzip"
        f = h5py.File(fullfilename, 'w')
        f.create_dataset("mol_name", data=self.mol)
        f.create_dataset("p", data=self.pgrid,compression=compression)
        f["p"].attrs["units"] = self.p_unit
        f.create_dataset("t", data=self.tgrid,compression=compression)
        f.create_dataset("xsecarr", data=self.kdata,compression=compression)
        f["xsecarr"].attrs["units"] = self.kdata_unit
        f.create_dataset("bin_edges", data=self.wns,compression=compression)
        f["bin_edges"].attrs["units"] = self.wn_unit
        f.close()

    def read_exo_transmit(self, filename, mol=None):
        """Creates an xsec object from an exo_transmit like spectra.
        See https://github.com/elizakempton/Exo_Transmit or Kempton et al. (2016) for details.
        Pressures are expected to be in Pa and cross sections in m^2/molecule

        Parameters
        ----------
            filename : str
                Name of the input file.
            mol: str
                Overrides the name of the molecule to be put in the :class:`Xtable` object.
        """
        tmp_wlgrid=[]
        tmp_kdata=[]
        with open(filename, 'r') as file:
            tmp = file.readline().split()
            self.tgrid=np.array([float(ii) for ii in tmp])
            tmp = file.readline().split()
            self.pgrid=np.array([float(ii) for ii in tmp])
            self.logpgrid=np.log10(self.pgrid)
            self.Np=self.pgrid.size
            self.Nt=self.tgrid.size
            while True:
                line=file.readline()
                if line is None or line=='': break
                tmp_wlgrid.append(float(line.split()[0]))
                tmp_kdata.append([])
                for _ in range(self.Np):
                    tmp_kdata[-1].append(file.readline().split()[1:])
        tmp_wlgrid=np.array(tmp_wlgrid)
        self.Nw=tmp_wlgrid.size
        self.kdata=np.zeros((self.Np,self.Nt,self.Nw))
        for iP in range(self.Np):
            for iT in range(self.Nt):
                for iW in range(self.Nw):
                    self.kdata[iP,iT,iW]=tmp_kdata[self.Nw-iW-1][iP][iT]
        self.wns=0.01/tmp_wlgrid[::-1]
        self.wnedges=np.concatenate( \
            ([self.wns[0]],(self.wns[:-1]+self.wns[1:])*0.5,[self.wns[-1]]))
        self.p_unit='Pa' 
        self.kdata_unit='m^2/molecule'
        if mol is not None:
            self.mol=mol
        else:
            self.mol=os.path.basename(filename).split(self._settings._delimiter)[0]

    def hires_to_xtable(self, path=None, filename_grid=None,
        logpgrid=None, tgrid=None,
        write=0, mol=None,
        grid_p_unit='Pa', p_unit='unspecified',
        kdata_unit='unspecified', file_kdata_unit='unspecified',
        **kwargs):
        """Loads an `Xtable` from high-resolution spectra.

        .. warning::
            By default, log pressures are specified in Pa in logpgrid!!! If you want
            to use another unit, do not forget to specify it with the grid_p_unit keyword.

        see :func:`exo_k.ktable.Ktable.hires_to_ktable` method for details
        on the arguments and options.
        """        
        first=True

        if path is None: raise TypeError("You should provide an input hires_spectrum directory")
        self.filename=path
        if mol is not None:
            self.mol=mol
        else:
            self.mol=os.path.basename(self.filename).split(self._settings._delimiter)[0]

        conversion_factor=u.Unit(grid_p_unit).to(u.Unit('Pa'))
        self.logpgrid=np.array(logpgrid, dtype=float)+np.log10(conversion_factor)
        self.pgrid=10**self.logpgrid #in Pa
        self.p_unit='Pa'
        self.Np=self.logpgrid.size
        if write >= 3 : print(self.Np,self.pgrid)

        self.tgrid=np.array(tgrid)
        self.Nt=self.tgrid.size
        if write >= 3 : print(self.Nt,self.tgrid)
        
        if filename_grid is None:
            filename_grid=create_fname_grid_Kspectrum_LMDZ(self.Np, self.Nt,
                **select_kwargs(kwargs,['suffix','nb_digit']))
        else:
            filename_grid=np.array(filename_grid)

        for iP in range(self.Np):
          for iT in range(self.Nt):
            filename=filename_grid[iP,iT]
            fname=os.path.join(path,filename)
            if write >= 3 : print(fname)

            spec_hr=Hires_spectrum(fname, file_kdata_unit=file_kdata_unit,
                **select_kwargs(kwargs,['skiprows','wn_column','mult_factor',
                    'kdata_column','data_type']))
            # for later conversion, the real kdata_unit is in spec_hr.kdata_unit
            self.kdata_unit=spec_hr.kdata_unit
            was_xsec=(spec_hr.data_type=='xsec')
            wn_hr=spec_hr.wns
            k_hr=spec_hr.kdata
            if not was_xsec:
                k_hr=k_hr*KBOLTZ*self.tgrid[iT]/self.pgrid[iP]
            if first:
                self.wns=wn_hr[1:-1]  #to be consistent with kcorr
                self.Nw=self.wns.size
                self.kdata=np.zeros((self.Np,self.Nt,self.Nw))
                self.kdata[iP,iT]=k_hr[1:-1]
                first=False
            else:
                self.kdata[iP,iT]=np.interp(self.wns,wn_hr[1:-1],k_hr[1:-1])
        if not was_xsec:
            self.kdata=self.kdata*u.Unit(self.kdata_unit).to(u.Unit('m^-1'))
            self.kdata_unit='m^2/molecule'
            # Accounts for the conversion of the abs_coeff to m^-1, so we know that
            #  self.kdata is now in m^2/molecule. Can now convert to the desired unit.
        if self._settings._convert_to_mks and kdata_unit is 'unspecified':
            kdata_unit='m^2/molecule'
        self.convert_kdata_unit(kdata_unit=kdata_unit)
        # converts from self.kdata_unit wich is either:
        #   - 'm^2/molecule' data_type was 'abs_coeff'
        #   - The units after Hires_spectrum (which have already taken
        #        file_kdata_unit into account)
        #  to the desired unit.
        self.convert_p_unit(p_unit=p_unit)

    def bin_down(self, wnedges=None, remove_zeros=False, write=0):
        """Method to bin down a xsec table to a new grid of wavenumbers (in place)

        Parameters
        ----------
            wnedges : array
                Edges of the new bins of wavenumbers (cm-1) onto which the xsec
                should be binned down.
                if you want Nwnew bin in the end, wngrid.size must be Nwnew+1
                wnedges[0] should be greater than self.wnedges[0]
                wnedges[-1] should be lower than self.wnedges[-1]
            remove_zeros: bool, optional
                If True, remove zeros in kdata. 
        """
        wnedges=np.array(wnedges)
        if wnedges.size<2: raise TypeError('wnedges should at least have two values')
        wngrid_filter = np.where((wnedges <= self.wnedges[-1]) & (wnedges >= self.wnedges[0]))[0]
        if write>=3:
            print(self.wnedges);print(wnedges);print(wngrid_filter);print(wnedges[wngrid_filter])

        indicestosum,weights=rebin_ind_weights(self.wnedges, wnedges[wngrid_filter])
        if write>=3: print(indicestosum[0]);print(weights[0])
        Nnew=wnedges.size-1
        Ntmp=wnedges[wngrid_filter].size-1
        newxsec=np.zeros((self.Np,self.Nt,Nnew))
        for iP in range(self.Np):
            for iT in range(self.Nt):
                tmp=self.kdata[iP,iT,:]
                newxsec[iP,iT,wngrid_filter[0:-1]]= \
                    [np.dot(tmp[indicestosum[ii]-1:indicestosum[ii+1]],weights[ii]) \
                        for ii in range(Ntmp)]        
        self.kdata=newxsec
        self.wnedges=wnedges
        self.wns=(wnedges[1:]+wnedges[:-1])*0.5
        self.Nw=self.wns.size
        if remove_zeros : self.remove_zeros(deltalog_min_value=10.)

    def sample(self, wngrid, remove_zeros=False, log_interp=None):
        """Method to re sample a xsec table to a new grid of wavenumbers (in place)

        Parameters
        ----------
            wngrid : array
                Location of the new wavenumbers points (cm-1)
        """
        wngrid=np.array(wngrid)
        wngrid_filter = np.where((wngrid <= self.wnedges[-1]) & (wngrid >= self.wnedges[0]))[0]
        Nnew=wngrid.size
        newxsec=np.zeros((self.Np,self.Nt,Nnew))
        if log_interp is None: log_interp=self._settings._log_interp
        if log_interp:
            for iP in range(self.Np):
                for iT in range(self.Nt):
                    tmp=np.log(self.kdata[iP,iT,:])
                    newxsec[iP,iT,wngrid_filter]=np.interp(wngrid[wngrid_filter],self.wns,tmp)
            self.kdata=np.exp(newxsec)
        else:
            for iP in range(self.Np):
                for iT in range(self.Nt):
                    tmp=self.kdata[iP,iT,:]
                    newxsec[iP,iT,wngrid_filter]=np.interp(wngrid[wngrid_filter],self.wns,tmp)
            self.kdata=newxsec
        self.wns=wngrid
        self.wnedges=np.concatenate( \
            ([self.wns[0]],0.5*(self.wns[1:]+self.wns[:-1]),[self.wns[-1]]))
        self.Nw=Nnew
        if remove_zeros : self.remove_zeros(deltalog_min_value=10.)

    def sample_cp(self, wngrid, **kwargs):
        """Creates a copy of the instance before resampling it.

        Parameters
        ----------
            See sample method for details. 

        Returns
        -------
            :class:`Xtable` object
                the re-sampled :class:`Xtable`
        """
        res=self.copy()
        res.sample(wngrid, **kwargs)
        return res


    def spectrum_to_plot(self, p=1.e-5, t=200., x=1., g=None):
        """provide the spectrum for a given point to be plotted

        Parameters
        ----------
            p : float
                Pressure (Ktable pressure unit)
            t : float
                Temperature(K)
            x: float
                Volume mixing ratio of the species
            g: is unused but here to be consistent with the method in data_table
        """
        return self.interpolate_kdata(log10(p),t)[0]*x


    def copy(self, cp_kdata=True):
        """Creates a new instance of :class:`Xtable` object and (deep) copies data into it

        Parameters
        ----------
            cp_kdata: bool, optional
                If false, the kdata table is not copied and
                only the structure and metadata are. 

        Returns
        -------
            :class:`Xtable`
                A new :class:`Xtable` instance with the same structure as self.
        """
        res=Xtable()
        res.copy_attr(self, cp_kdata=cp_kdata)
        return res

    def __repr__(self):
        """Method to output header
        """
        out1=super().__repr__()
        output=out1+"""
        data oredered following p, t, wn
        shape        : {shape}
        """.format(shape=self.shape)
        return output



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
Doc can be created with "pydoc -w corrk_lib"
"""
import os
from math import log10
import time
import pickle
import h5py
import numpy as np
import astropy.units as u
from .data_table import Data_table
from .ktable5d import Ktable5d
from .util.interp import rm_molec, interp_ind_weights, rebin_ind_weights, rebin, is_sorted, \
        gauss_legendre, spectrum_to_kdist, kdata_conv_loop, bin_down_corrk_numba
from .hires_spectrum import Hires_spectrum
from .util.filenames import create_fname_grid_Kspectrum_LMDZ, select_kwargs, read_nemesis_binary
from .util.cst import KBOLTZ, nemesis_hitran_id_numbers


class Ktable(Data_table):
    """A class that handles 4D tables of k-coefficients.
    Based on the Data_table class that handles basic operations common to Xtable.
    """
        
    def __init__(self, *filename_filters, filename=None, xtable=None, path=None,
        p_unit='unspecified', file_p_unit='unspecified',
        kdata_unit='unspecified', file_kdata_unit='unspecified',
        remove_zeros=False, search_path=None, mol=None,
        **kwargs):
        """Initializes k coeff table and supporting data from various sources
        (see below by order of precedence)

        Parameters
        ----------
            filename: str, optional
                Relative or absolute name of the file to be loaded. 
            filename_filters: sequence of string
                As many strings as necessary to uniquely define
                a file in the global search path defined in
                :class:`~exo_k.settings.Settings`.
                This path will be searched for a file
                with all the filename_filters in the name.
                The filename_filters can contain '*'.
            xtable: Xtable object
                If no filename nor filename_filters are provided, this xtable object will be used to
                create a ktable. In this case, wavenumber bins must be given
                with the wnedges keyword.
            path: str
                If none of the above is specifed, path can point to
                a directory with a LMDZ type k coeff table.
                In this case, see read_LMDZ for the keywords to specify.

        If none of the parameters above is specified,
        just creates an empty object to be filled later.

        Other Parameters
        ----------------
            p_unit: str, optional
                String identifying the pressure units to convert to (e.g. 'bar', 'Pa', 'mbar', 
                or any pressure unit recognized by the astropy.units library).
                If ='unspecified', no conversion is done.
            file_p_unit: str, optional
                String to specify the current pressure unit if it is unspecified or if 
                you have reasons to believe it is wrong (e.g. you just read a file where
                you know that the pressure grid and the pressure unit do not correspond)
            kdata_unit: str, optional
                String to identify the unit to convert to.
                Accepts 'cm^2', 'm^2'
                or any surface unit recognized by the astropy.units library.
                If ='unspecified', no conversion is done.
                In general, kdata should be kept in 'per number' or 'per volume'
                units (as opposed to 'per mass' units) as composition will
                always be assumed to be a number or volume mixing ratio.
                Opacities per unit mass are not supported yet.
                Note that you do not need to specify the '/molec' or '/molecule' in the unit.
            file_kdata_unit: str, optional
                String to specify the current kdata unit if it is unspecified or if 
                you have reasons to believe it is wrong (e.g. you just read a file where
                you know that the kdata grid and the kdata unit do not correspond)
            remove_zeros: boolean, optional
                If True, the zeros in the kdata table are replaced by
                a value 10 orders of magnitude smaller than the smallest positive value
            mol: str, optional
                The name of the gas or molecule described by the :class:`Ktable`
            search_path: str, optional
                If search_path is provided,
                it locally overrides the global _search_path
                in :class:`~exo_k.settings.Settings`
                and only files in search_path are returned.
        """
        super().__init__()

        if filename is not None:
            self.filename=filename
        elif filename_filters or (mol is not None and (xtable is None and path is None)): 
        # a none empty sequence returns a True in a conditional statement
            self.filename=self._settings.list_files(*filename_filters, molecule = mol,
                only_one=True, search_path=search_path)[0]

        if self.filename is not None:
            if self.filename.lower().endswith('pickle'):
                self.read_pickle(filename=self.filename, mol=mol)
            elif self.filename.lower().endswith(('.hdf5', '.h5')):
                self.read_hdf5(filename=self.filename, mol=mol)
            elif self.filename.lower().endswith('.kta'):
                self.read_nemesis(filename=self.filename, mol=mol)
            else:
                raise NotImplementedError("""Requested format not recognized.
            Should end with .pickle, .hdf5, .h5, or .kta""")
        elif xtable is not None:
            self.xtable_to_ktable(xtable=xtable, **kwargs)
        elif path is not None:
            self.read_LMDZ(path=path, mol=mol, **kwargs)
        else:                  #if there is no input file, just create an empty object 
            self.wnedges=None
            self.weights=None
            self.ggrid=None
            self.gedges=None

        super().finalize_init(p_unit=p_unit, file_p_unit=file_p_unit,
            kdata_unit=kdata_unit, file_kdata_unit=file_kdata_unit,
            remove_zeros=remove_zeros)


    @property
    def shape(self):
        """Returns the shape of self.kdata
        """
        return np.array([self.Np,self.Nt,self.Nw,self.Ng])

    def read_pickle(self, filename=None, mol=None):
        """Initializes k coeff table and supporting data from an Exomol pickle file

        Parameters
        ----------
            filename: str
                Name of the input pickle file
            mol: str, optional
                Force the name of the molecule
        """
        if filename is None: raise RuntimeError("You should provide an input pickle filename")
        pickle_file=open(filename,'rb')
        raw=pickle.load(pickle_file, encoding='latin1')
        pickle_file.close()
        
        if mol is not None: 
            self.mol=mol
        else:
            self.mol=raw['name']
            if self.mol=='H2OP': self.mol='H2O'

        self.pgrid=raw['p']
        self.logpgrid=np.log10(self.pgrid)
        self.tgrid=raw['t']
        self.wns=raw['bin_centers']
        self.wnedges=raw['bin_edges']
        if 'weights' in raw.keys():
            self.weights=raw['weights']
        else:
            raise RuntimeError('No weights keyword. This file is probably a cross section file.')
        self.ggrid=raw['samples']
        self.gedges=np.insert(np.cumsum(self.weights),0,0.)
        self.kdata=raw['kcoeff']
    
        if 'p_unit' in raw.keys():
            self.p_unit=raw['p_unit']
        else:
            self.p_unit='bar'

        if 'kdata_unit' in raw.keys():
            self.kdata_unit=raw['kdata_unit']
        else:
            self.kdata_unit='cm^2/molec'

        if 'wn_unit' in raw.keys(): self.wn_unit=raw['wn_unit']

        self.Np,self.Nt,self.Nw,self.Ng=self.kdata.shape

    def write_pickle(self, filename):
        """Saves data in a pickle format

        Parameters
        ----------
            filename: str
                Name of the file to be created and saved
        """
        fullfilename=filename
        if not filename.lower().endswith('.pickle'): fullfilename=filename+'.pickle'
        pickle_file=open(fullfilename,'wb')
        dictout={'name':self.mol,
                 'p':self.pgrid,
                 'p_unit':self.p_unit,
                 't':self.tgrid,
                 'bin_centers':self.wns,
                 'bin_edges':self.wnedges,
                 'wn_unit':self.wn_unit,
                 'weights':self.weights,
                 'samples':self.ggrid,
                 'kcoeff':self.kdata,
                 'kdata_unit':self.kdata_unit}
        #print(dictout)
        pickle.dump(dictout,pickle_file,protocol=-1)
        pickle_file.close()

    def read_hdf5(self, filename=None, mol=None):
        """Initializes k coeff table and supporting data from an Exomol hdf5 file

        Parameters
        ----------
            file : str
                Name of the input hdf5 file
        """
        if (filename is None or not filename.lower().endswith(('.hdf5', '.h5'))):
            raise RuntimeError("You should provide an input hdf5 file")
        f = h5py.File(filename, 'r')
        if 'mol_name' in f:
            self.mol=f['mol_name'][()]
        elif 'mol_name' in f.attrs:
            self.mol=f.attrs['mol_name']
        else:
            self.mol=os.path.basename(filename).split(self._settings._delimiter)[0]
        if isinstance(self.mol, np.ndarray): self.mol=self.mol[0]
        if mol is not None:
            self.mol=mol
        self.wns=f['bin_centers'][...]
        self.wnedges=f['bin_edges'][...]
        if 'units' in f['bin_edges'].attrs:
            self.wn_unit=f['bin_edges'].attrs['units']
        else:
            if 'units' in f['bin_centers'].attrs:
                self.wn_unit=f['bin_centers'].attrs['units']
        self.kdata=f['kcoeff'][...]
        self.kdata_unit=f['kcoeff'].attrs['units']
        self.tgrid=f['t'][...]
        self.pgrid=f['p'][...]
        self.logpgrid=np.log10(self.pgrid)
        self.p_unit=f['p'].attrs['units']
        if 'weights' in f.keys():
            self.weights=f['weights'][...]
        else:
            raise RuntimeError('No weights keyword. This file is probably a cross section file.')
        self.ggrid=f['samples'][...]
        self.gedges=np.insert(np.cumsum(self.weights),0,0.)
        self.logk=False
        f.close()  
        self.Np,self.Nt,self.Nw,self.Ng=self.kdata.shape

    def write_hdf5(self, filename):
        """Saves data in a hdf5 format

        Parameters
        ----------
            filename: str
                Name of the file to be created and saved
        """
        fullfilename=filename
        if not filename.lower().endswith(('.hdf5', '.h5')):
            fullfilename=filename+'.hdf5'
        compression="gzip"
        f = h5py.File(fullfilename, 'w')
        f.create_dataset("mol_name", data=self.mol)
        f.create_dataset("p", data=self.pgrid,compression=compression)
        f["p"].attrs["units"] = self.p_unit
        f.create_dataset("t", data=self.tgrid,compression=compression)
        f.create_dataset("kcoeff", data=self.kdata,compression=compression)
        f["kcoeff"].attrs["units"] = self.kdata_unit
        f.create_dataset("samples", data=self.ggrid,compression=compression)
        f.create_dataset("weights", data=self.weights,compression=compression)
        f.create_dataset("bin_edges", data=self.wnedges,compression=compression)
        f.create_dataset("bin_centers", data=self.wns,compression=compression)
        f["bin_edges"].attrs["units"] = self.wn_unit
        f.close()    

    def read_LMDZ(self, path=None, res=None, band=None, mol=None):
        """Initializes k coeff table and supporting data from a .dat file in a gcm friendly format.
        
        Units are assumed to be cm^2 for kdata and mbar for pressure. 

        Parameters
        ----------
            path: str
                Name of the directory with the various input files
            res: str
                "IRxVI" where IR and VI are the numbers of bands
                in the infrared and visible of the k table to load.
            band: str
                "IR" or "VI" to specify which band to load.
            mol: str
                Name of the molecule to be saved in the Ktable object. 
        """        
        if (path is None) or (res is None): \
            raise TypeError("You should provide an input directory name and a resolution")

        self.filename=path
        if mol is not None:
            self.mol=mol
        else:
            self.mol=os.path.basename(self.filename).split(self._settings._delimiter)[0]

        self.weights=np.loadtxt(os.path.join(path,'g.dat'),skiprows=1)[:-1]
            # we remove the last point that is always zero.
            # in the gcm this last point is intended to take care of continuum
        self.Ng=self.weights.size
        self.gedges=np.insert(np.cumsum(self.weights),0,0.)
        self.ggrid=(self.gedges[1:]+self.gedges[:-1])*0.5

        self.p_unit='mbar'
        self.logpgrid=np.loadtxt(os.path.join(path,'p.dat'),skiprows=1)*1.
        self.Np=self.logpgrid.size
        self.pgrid=10**self.logpgrid

        self.tgrid=np.loadtxt(os.path.join(path,'T.dat'),skiprows=1)
        self.Nt=self.tgrid.size

        if band is None:
            raw=np.loadtxt(os.path.join(path,res,'narrowbands.in'), skiprows=1, unpack=True)
        else:
            raw=np.loadtxt(os.path.join(path,res,'narrowbands_'+band+'.in'), \
                skiprows=1, unpack=True)
        self.wnedges=np.append(raw[0],raw[1,-1])
        self.wns=(self.wnedges[1:]+self.wnedges[:-1])*0.5
        self.Nw=self.wns.size
        
        self.kdata_unit='cm^2/molecule'
        if band is None:
            file_to_load=os.path.join(path,res,'corrk_gcm.dat')
        else:
            file_to_load=os.path.join(path,res,'corrk_gcm_'+band+'.dat')        
        tmp=np.loadtxt(file_to_load).reshape((self.Nt,self.Np,self.Nw,self.Ng+1),order='F')
        self.kdata=tmp[:,:,:,:-1].transpose((1,0,2,3)) 
        # also removing the last g point which is equal to 0.
        self.logk=False        
        return None

    def write_LMDZ(self, path, band='IR', fmt='%22.15e', write_only_metadata=False):
        """Saves data in a LMDZ friendly format.
        
        The gcm requires p in mbar and kdata in cm^2/molec.
        The conversion is done automatically.

        Parameters
        ----------
            path: str
                Name of the directory to be created and saved,
                the one that will contain all the necessary files
            band: str
                The band you are computing: 'IR' or 'VI'
            fmt: str
                Fortran format for the corrk file. 
        """
        try:
            os.mkdir(path)
        except FileExistsError:
            print('Directory was already there '+path)
        file = open(os.path.join(path,'p.dat'), "w")
        file.write(str(self.Np)+'\n')
        lp_to_write=self.logpgrid+np.log10(u.Unit(self.p_unit).to(u.Unit('mbar')))
        for lp in lp_to_write:
            file.write(str(lp)+'\n')
        file.close()

        file = open(os.path.join(path,'T.dat'), "w")
        file.write(str(self.Nt)+'\n')
        for t in self.tgrid:
            file.write(str(t)+'\n')
        file.close()

        file = open(os.path.join(path,'g.dat'), "w")
        file.write(str(self.Ng+1)+'\n')
        for g in self.weights:
            file.write(str(g)+'\n')
        file.write(str(0.)+'\n')
        file.close()

        dirname=os.path.join(path,band+str(self.Nw))
        try:
            os.mkdir(dirname)
        except FileExistsError:
            print('Directory was already there: '+dirname)

        file = open(os.path.join(dirname,'narrowbands_'+band+'.in'), "w")
        file.write(str(self.Nw)+'\n')
        for iw in range(self.Nw):
            file.write(str(self.wnedges[iw])+' '+str(self.wnedges[iw+1])+'\n')
        file.close()

        if not write_only_metadata:
            #file = open(dirname+'/corrk_gcm_IR.in', "w")
            data_to_write=self.kdata.transpose((1,0,2,3)).flatten(order='F')
            data_to_write=data_to_write*u.Unit(rm_molec(self.kdata_unit)).to(u.Unit('cm^2'))
            data_to_write=np.append(data_to_write, \
                np.zeros(self.Np*self.Nt*self.Nw)).reshape((1,self.Np*self.Nt*self.Nw*(self.Ng+1)))
            np.savetxt(os.path.join(dirname,'corrk_gcm_'+band+'.dat'),data_to_write,fmt=fmt)

    def read_nemesis(self, filename=None, mol=None):
        """Initializes k coeff table and supporting data from a Nemesis binary file (.kta)

        Parameters
        ----------
            file: str
                Name of the input Nemesis binary file.
            mol: str, optional
                Name of the molecule.
        """
        if (filename is None or not filename.lower().endswith('.kta')):
            raise RuntimeError("You should provide an input nemesis (.kta) file")
        
        if mol is not None:
            self.mol=mol
        else:
            self.mol=os.path.basename(self.filename).split(self._settings._delimiter)[0]

        self.Np, self.Nt, self.Nw, self.Ng, \
            self.pgrid, self.tgrid, self.wns, \
            self.ggrid, self.weights, self.kdata = read_nemesis_binary(filename)

        self.logpgrid=np.log10(self.pgrid)
        self.wnedges=np.concatenate(  \
            ([self.wns[0]],(self.wns[:-1]+self.wns[1:])*0.5,[self.wns[-1]]))
        self.gedges=np.insert(np.cumsum(self.weights),0,0.)
        self.p_unit='bar'
        self.kdata_unit='cm^2/molecule'

    def write_nemesis(self, filename):
        """Saves data in a nemesis format.

        base on a routine provided by K. Chubb.

        Parameters
        ----------
            filename: str
                Name of the file to be created and saved
        """
        int_format = np.int32
        float_format = np.float32
        with open(filename,'wb') as o:
            o.write(int_format(11+2*self.Ng+2+self.Np+self.Nt+self.Nw).tostring())
            #int for irec0 (5221)
            o.write(int_format(self.Nw).tostring()) #int
            o.write(float_format(self.wls[-1]).tostring()) #float for VMIN (0.30015)
            o.write(float_format(-1.).tostring()) #float
            o.write(float_format(0.).tostring()) #float for FWHM
            o.write(int_format(self.Np).tostring())
            o.write(int_format(self.Nt).tostring())
            o.write(int_format(self.Ng).tostring())
            o.write(int_format(nemesis_hitran_id_numbers[self.mol]).tostring()) #IDGAS1 FROM HITRAN
            o.write(int_format(0).tostring())
            o.write(float_format(self.ggrid).tostring())
            o.write(float_format(self.weights).tostring())
            o.write(float_format(0.).tostring()) #float for 0
            o.write(float_format(0.).tostring()) #float for 0
            o.write(float_format(self.pgrid).tostring())
            o.write(float_format(self.tgrid).tostring())
            o.write(float_format(self.wls[::-1]).tostring())
            o.write(float_format(self.kdata[:,:,::-1,:].transpose(2,0,1,3)*1.e20).tostring())

    def xtable_to_ktable(self, xtable=None, wnedges=None, weights=None, ggrid=None,
        quad='legendre', order=20, mid_dw=True, write=0):
        """Fills the :class:`~exo_k.ktable.Ktable` object with a k-coeff table computed
        from a :class:`~exo_k.xtable.Xtable` object.

        The p and kcorr units are inherited from the :class:`~exo_k.xtable.Xtable` object.

        Parameters
        ----------
            xtable: :class:`~exo_k.xtable.Xtable`
                input Xtable object instance
            wnedges: Array
                edges of the wavenumber bins to be used to compute the corrk

        Other Parameters
        ----------------
            weights: array, optional
                If weights are provided, they are used instead of the legendre quadrature. 
            quad: string, optional
                Type of quadrature used. Default is 'legendre'
            order: Integer, optional
                Order of the Gauss legendre quadrature used. Default is 20.
            mid_dw: boolean, optional

                * If True, the Xsec values in the high resolution xtable data are assumed to
                  cover a spectral interval that is centered around
                  the corresponding wavenumber value.
                  The first and last Xsec values are discarded. 
                * If False, each interval runs from the wavenumber value to the next one.
                  The last Xsec value is dicarded.
        """        
        if xtable is None: raise TypeError("You should provide an input Xtable object")
        if wnedges is None: raise TypeError("You should provide an input wavenumber array")

        self.copy_attr(xtable,cp_kdata=False)
        self.wnedges=np.array(wnedges)
        if self.wnedges.size<2: raise TypeError('wnedges should at least have two values')
        self.wns=0.5*(self.wnedges[1:]+self.wnedges[:-1])

        if weights is not None:
            self.weights=np.array(weights)
            self.gedges=np.concatenate(([0],np.cumsum(self.weights)))
            if ggrid is not None: 
                self.ggrid=np.array(ggrid)
            else:
                self.ggrid=(self.gedges[1:]+self.gedges[:-1])*0.5
        else:
            if quad=='legendre':
                self.weights,self.ggrid,self.gedges=gauss_legendre(order)
            else:
                raise NotImplementedError("Type of quadrature (quad keyword) not known.")
        self.Ng=self.weights.size
        if write >6 : print(self.weights,self.ggrid,self.gedges)
        self.logk=False
        self.Nw=self.wns.size
        self.kdata=np.zeros(self.shape)
        if mid_dw:
            dwn_hr=(xtable.wns[2:]-xtable.wns[:-2])*0.5
            wn_hr=xtable.wns[1:-1]
        else:
            dwn_hr=(xtable.wns[1:]-xtable.wns[:-1])
            wn_hr=xtable.wns[:-1]
        for iP in range(self.Np):
          for iT in range(self.Nt):
            if mid_dw:
                k_hr=xtable.kdata[iP,iT,1:-1]
            else:
                k_hr=xtable.kdata[iP,iT,:-1]
            #print(self.ggrid)
            #print(wn_hr[[0,-1]])
            #print(self.wnedges[[0,-1]])
            self.kdata[iP,iT]=spectrum_to_kdist(k_hr,wn_hr,dwn_hr,self.wnedges,self.ggrid)

    def hires_to_ktable(self, path=None, filename_grid=None,
        logpgrid=None, tgrid=None, wnedges=None,
        quad='legendre', order=20, weights=None, ggrid=None,
        mid_dw=True, write=0, mol=None,
        grid_p_unit='Pa', p_unit='unspecified',
        kdata_unit='unspecified', file_kdata_unit='unspecified',
        **kwargs):
        """Computes a k coeff table from :class:`~exo_k.util.hires_spectrum.Hires_spectrum`
        objects.

        .. warning::
            By default, log pressures are specified in Pa in logpgrid!!! If you want
            to use another unit, do not forget to specify it with the grid_p_unit keyword.

        Parameters
        ----------
            path : String
                directory with the input files
            filename_grid : array of str with shape (logpgrid.size,tgrid.size)
                Names of the input high-res spectra. If None, the files are assumed to
                follow Kspectrum/LMDZ convention, i.e.
                be of the type 'k001', 'k002', etc. 
                See :func:`~exo_k.util.filenames.create_fname_grid_Kspectrum_LMDZ`
                for possible additional keyword arguments. 
            logpgrid: array
                Grid in log(pressure) of the input. Default unit is Pa, but can be changed
                with the `grid_p_unit` keyword.
            tgrid: array
                Grid in temperature of the input.
            wnedges : array
                Edges of the wavenumber bins to be used to compute the corr-k

        Other Parameters
        ----------------
            weights: array, optional
                If weights are provided, they are used instead of the legendre quadrature. 
            quad : string, optional
                Type of quadrature used. Default is 'legendre'
            order : Integer, optional
                Order of the Gauss legendre quadrature used. Default is 20.
            mid_dw: boolean, optional
                * If True, the Xsec values in the high resolution xsec data are assumed to
                  cover a spectral interval that is centered around
                  the corresponding wavenumber value.
                  The first and last Xsec values are discarded. 
                * If False, each interval runs from the wavenumber value to the next one.
                  The last Xsec value is dicarded.
            mol: string, optional
                Give a name to the molecule. Useful when used later in a Kdatabase
                to track molecules.
            p_unit: str, optional
                Pressure unit to convert to.
            grid_p_unit : str, optional
                Unit of the specified `logpgrid`.
            kdata_unit : str, optional
                Kdata unit to convert to.
            file_kdata_unit : str, optional
                Kdata unit in input files.

        See :func:`~exo_k.util.filenames.create_fname_grid_Kspectrum_LMDZ`
        or :class:`~exo_k.util.hires_spectrum.Hires_spectrum`
        for a list of additional arguments that can be provided to those funtion
        through `**kwargs`.
        """        
        if path is None: raise TypeError("You should provide an input hires_spectrum directory")
        if wnedges is None: raise TypeError("You should provide an input wavenumber array")

        self.filename=path
        if mol is not None:
            self.mol=mol
        else:
            self.mol=os.path.basename(self.filename).split(self._settings._delimiter)[0]

        if weights is not None:
            self.weights=np.array(weights)
            self.gedges=np.concatenate(([0],np.cumsum(self.weights)))
            if ggrid is not None: 
                self.ggrid=np.array(ggrid)
            else:
                self.ggrid=(self.gedges[1:]+self.gedges[:-1])*0.5
        else:
            if quad=='legendre':
                self.weights,self.ggrid,self.gedges=gauss_legendre(order)
            else:
                raise NotImplementedError("Type of quadrature (quad keyword) not known.")
        self.Ng=self.weights.size

        conversion_factor=u.Unit(grid_p_unit).to(u.Unit('Pa'))
        self.logpgrid=np.array(logpgrid, dtype=float)+np.log10(conversion_factor)
        self.pgrid=10**self.logpgrid #in Pa
        self.p_unit='Pa'
        self.Np=self.logpgrid.size
        if write >= 3 : print(self.Np,self.pgrid)

        self.tgrid=np.array(tgrid)
        self.Nt=self.tgrid.size
        if write >= 3 : print(self.Nt,self.tgrid)

        self.wnedges=np.array(wnedges)
        if self.wnedges.size<2: raise TypeError('wnedges should at least have two values')
        self.wns=(self.wnedges[1:]+self.wnedges[:-1])*0.5
        self.Nw=self.wns.size
        
        self.kdata=np.zeros(self.shape)
        if filename_grid is None:
            filename_grid=create_fname_grid_Kspectrum_LMDZ(self.Np, self.Nt,
                **select_kwargs(kwargs,['suffix','nb_digit']))
        else:
            filename_grid=np.array(filename_grid)
        
        for iP in range(self.Np):
          for iT in range(self.Nt):
            filename=filename_grid[iP,iT]
            fname=os.path.join(path,filename)
            if write >= 3 : print(fname)

            spec_hr=Hires_spectrum(fname, file_kdata_unit=file_kdata_unit,
                **select_kwargs(kwargs,['skiprows','wn_column','mult_factor',
                    'kdata_column','data_type']))
            # for later conversion, the real kdata_unit is in spec_hr.kdata_unit
            self.kdata_unit=spec_hr.kdata_unit
            was_xsec=(spec_hr.data_type=='xsec')
            wn_hr=spec_hr.wns
            k_hr=spec_hr.kdata
            if mid_dw:
                dwn_hr=(wn_hr[2:]-wn_hr[:-2])*0.5
                wn_hr=wn_hr[1:-1]
                k_hr=k_hr[1:-1]
            else:
                dwn_hr=(wn_hr[1:]-wn_hr[:-1])
                wn_hr=wn_hr[:-1]
                k_hr=k_hr[:-1]
            self.kdata[iP,iT]=spectrum_to_kdist(k_hr,wn_hr,dwn_hr,self.wnedges,self.ggrid)
            if not was_xsec:
                self.kdata[iP,iT]=self.kdata[iP,iT]*KBOLTZ*self.tgrid[iT]/self.pgrid[iP]
        if not was_xsec:
            self.kdata=self.kdata*u.Unit(self.kdata_unit).to(u.Unit('m^-1'))
            self.kdata_unit='m^2/molecule'
            # Accounts for the conversion of the abs_coeff to m^-1, so we know that
            #  self.kdata is now in m^2/molecule. Can now convert to the desired unit.
        if self._settings._convert_to_mks and kdata_unit is 'unspecified':
            kdata_unit='m^2/molecule'
        self.convert_kdata_unit(kdata_unit=kdata_unit)
        # converts from self.kdata_unit wich is either:
        #   - 'm^2/molecule' data_type was 'abs_coeff'
        #   - The units after Hires_spectrum (which have already taken
        #        file_kdata_unit into account)
        #  to the desired unit.
        self.convert_p_unit(p_unit=p_unit)


    def copy(self, cp_kdata=True, ktab5d=False):
        """Creates a new instance of :class:`Ktable` object and (deep) copies data into it

        Parameters
        ----------
            cp_kdata: bool, optional
                If false, the kdata table is not copied and
                only the structure and metadata are. 
            ktab5d: bool, optional
                If true, creates a Ktable5d object with the same structure. 
                Data are not copied.

        Returns
        -------
            :class:`Ktable` or :class:`Ktable5d`
                A new :class:`Ktable` or :class:`Ktable5d` 
                instance with the same structure as self.
        """
        if ktab5d:
            res=Ktable5d()
            cp_kdata=False
        else:    
            res=Ktable()
        res.copy_attr(self,cp_kdata=cp_kdata)
        res.weights = np.copy(self.weights)
        res.ggrid   = np.copy(self.ggrid)
        res.gedges  = np.copy(self.gedges)
        return res

    def gindex(self, g):
        """Finds the index corresponding to the given g
        """
        return min(np.searchsorted(self.ggrid,g),self.Ng-1)

    def spectrum_to_plot(self, p=1.e-5, t=200., x=1., g=None):
        """provide the spectrum for a given point to be plotted

        Parameters
        ----------
            p : float
                Pressure (Ktable pressure unit)
            t : float
                Temperature(K)
            g: float
                Gauss point
            x: float
                Mixing ratio of the species
        """
        if g is None: raise RuntimeError('A gauss point should be provided with the g= keyword.')
        gindex=self.gindex(g)
        return self.interpolate_kdata(log10(p), t, x)[0,:,gindex]

    def plot_distrib(self, ax, p=1.e-5, t=200., wl=1., x=1., xscale=None, yscale='log', **kwarg):
        """Plot the distribution for a given point

        Parameters
        ----------
            p : float
                Pressure (Ktable pressure unit)
            t : float
                Temperature(K)
            wl: float
                Wavelength (micron)
        """
        wlindex=self.wlindex(wl)
        toplot=self.interpolate_kdata(log10(p),t)[0,wlindex]*x
        if xscale is not None: 
            ax.set_xscale(xscale)
            ax.plot(1.-self.ggrid,toplot,**kwarg)
            ax.set_xlabel('1-g')
        else:
            ax.plot(self.ggrid,toplot,**kwarg)
            ax.set_xlabel('g')
        ax.set_ylabel('Cross section ('+self.kdata_unit+')')
        ax.grid(True)
        if yscale is not None: ax.set_yscale(yscale)

    def __repr__(self):
        """Method to output header
        """
        out1=super().__repr__()
        output=out1+"""
        weights      : {wg}
        data oredered following p, t, wn, g
        shape        : {shape}
        """.format(wg=self.weights, shape=self.shape)
        return output

    def RandOverlap(self, other, x_self, x_other, write=0, use_rebin=False):
        """Method to randomly mix the opacities of 2 species (self and other).

        Parameters
        ----------
            other: :class:`Ktable`
                A :class:`Ktable` object to be mixed with. Dimensions should be the same as self.
            x_self: float or array
                Volume mixing ratio of the first species
            x_other: float or array
                Volume mixing ratio of the species to be mixed with.

        If one of these is None, the kcoeffs of the species in question
        are considered to be already normalized with respect to the mixing ratio.

        Returns
        -------
            array
                array of k-coefficients for the mix. 
        """
        if not np.array_equal(self.shape,other.shape):
            raise TypeError("""in RandOverlap: kdata tables do not have the same dimensions.
                I'll stop now!""")
        Ng=self.Ng
        weights=self.weights
        try:
            kdatas=self.vmr_normalize(x_self)
        except TypeError:
            raise TypeError('Gave bad mixing ratio format to vmr_normalize')
        try:
            kdatao=other.vmr_normalize(x_other)
        except TypeError:
            raise TypeError('Gave bad mixing ratio format to vmr_normalize')
        kdataconv=np.zeros((self.Np,self.Nt,self.Nw,Ng**2))
        weightsconv=np.zeros(Ng**2)
        newkdata=np.zeros(self.shape)
#        newkdata2=np.zeros((self.shape[0]*self.shape[1]*self.shape[2],Ng))

        for jj in range(Ng):
            for ii in range(Ng):
                weightsconv[ii*Ng+jj]=weights[jj]*weights[ii]

        if write >= 3 : start=time.time()
        kdata_conv_loop(kdatas,kdatao,kdataconv,self.shape)
        if write >= 3 : end=time.time();print("kdata conv: ",end - start)        

        if write >= 3 : start=time.time()
        for iP in range(self.Np):
          for iT in range(self.Nt):
            for iW in range(self.Nw):
              tmp=kdataconv[iP,iT,iW,:]
              indices=np.argsort(tmp)
              kdatasort=tmp[indices]
              weightssort=weightsconv[indices]
              newggrid=np.cumsum(weightssort)
              if use_rebin:
                  newggrid=np.concatenate(([0.],newggrid))
                  kdatasort=np.concatenate(([kdatasort[0]],kdatasort))
                  newkdata[iP,iT,iW,:]=rebin(kdatasort,newggrid,self.gedges)
              else:
                  newkdata[iP,iT,iW,:]=np.interp(self.ggrid,newggrid,kdatasort)
        if write >= 3 : end=time.time();print("kdata rebin with loop: ",end - start)        

        return newkdata
        
    def bin_down(self, wnedges=None, weights=None, ggrid=None,
        remove_zeros=False, num=300, use_rebin=False, write=0):
        """Method to bin down a kcoeff table to a new grid of wavenumbers

        Parameters
        ----------
            wnedges: array
                Edges of the new bins of wavenumbers (cm-1)
                onto which the kcoeff should be binned down.
                if you want Nwnew bin in the end, wnedges.size must be Nwnew+1
                wnedges[0] should be greater than self.wnedges[0] (JL20 not sure anymore)
                wnedges[-1] should be lower than self.wnedges[-1]
            weights: array, optional
                Desired weights for the resulting Ktable.
            ggrid: array, optional
                Desired g-points for the resulting Ktable.
                Must be consistent with provided weights.
                If not given, they are taken at the midpoints of the array
                given by the cumulative sum of the weights
        """
        old_ggrid=self.ggrid
        if weights is not None:
            self.weights=np.array(weights)
            self.gedges=np.concatenate(([0],np.cumsum(self.weights)))
            if ggrid is not None: 
                self.ggrid=np.array(ggrid)
            else:
                self.ggrid=(self.gedges[1:]+self.gedges[:-1])*0.5
            self.Ng=self.ggrid.size
        wnedges=np.array(wnedges)
        if wnedges.size<2: raise TypeError('wnedges should at least have two values')
        wngrid_filter = np.where((wnedges <= self.wnedges[-1]) & (wnedges >= self.wnedges[0]))[0]
        if not is_sorted(wnedges):
            raise RuntimeError('wnedges should be sorted.')
        indicestosum,wn_weigths=rebin_ind_weights(self.wnedges, wnedges[wngrid_filter])
        if write> 10 :print(indicestosum);print(wn_weigths)

        self.kdata=bin_down_corrk_numba((self.Np,self.Nt,wnedges.size-1,self.Ng), \
            self.kdata, old_ggrid, self.ggrid, self.gedges, indicestosum, wngrid_filter, \
            wn_weigths, num, use_rebin)

        self.wnedges=wnedges
        self.wns=(wnedges[1:]+wnedges[:-1])*0.5
        self.Nw=self.wns.size
        if remove_zeros : self.remove_zeros(deltalog_min_value=10.)


########### OBSOLETE METHODS #################

    def remap_logPT_grid(self, logp_array=None, t_array=None):
        """Obsolete method. Replaced by remap_logPT in the parent class: Data_table.
        """
        res=Ktable()
        res.mol     =self.mol
        res.logpgrid=logp_array
        res.pgrid   =10**res.logpgrid
        res.tgrid   =t_array
        res.wns     =self.wns
        res.wnedges =self.wnedges
        res.weights =self.weights
        res.ggrid   =self.ggrid
        res.gedges  =self.gedges
        res.Np      =logp_array.shape
        res.Nt      =t_array.shape
        res.Nw      =self.Nw
        res.Ng      =self.Ng
        res.logk    =self.logk
        res.p_unit  =self.p_unit
        res.kdata_unit=self.kdata_unit
        #res.kdata=np.zeros((res.Np,res.Nt,res.Nw ,res.Ng))
        tind,tweight=interp_ind_weights(t_array,self.tgrid)
        lpind,lpweight=interp_ind_weights(logp_array,self.logpgrid)
        #print(tind,tweight)
        #print(lpind,lpweight)
        lpindextended=lpind.reshape((lpind.size,1))
        tw=tweight.reshape((1,tweight.size,1,1))    
        pw=lpweight.reshape((lpweight.size,1,1,1))
        # trick to broadcast over Nw and Ng a few lines below
        kc_p1t1=self.kdata[lpindextended,tind]
        kc_p0t1=self.kdata[lpindextended-1,tind]
        kc_p1t0=self.kdata[lpindextended,tind-1]
        kc_p0t0=self.kdata[lpindextended-1,tind-1]
        # kdata_tmp=kc_p1t1*pw*tw+kc_p0t1*(1.-pw)*tw+kc_p1t0*pw*(1.-tw)+kc_p0t0*(1.-pw)*(1.-tw)
        kdata_tmp=  np.log10(kc_p1t1)*pw*tw          \
                        +np.log10(kc_p0t1)*(1.-pw)*tw \
                        +np.log10(kc_p1t0)*pw*(1.-tw) \
                        +np.log10(kc_p0t0)*(1.-pw)*(1.-tw)
        res.kdata=10**kdata_tmp
        #kdata_tmp=  (kc_p1t1)*pw*tw          \
        #                +(kc_p0t1)*(1.-pw)*tw \
        #                +(kc_p1t0)*pw*(1.-tw) \
        #                +(kc_p0t0)*(1.-pw)*(1.-tw)
        #res.kdata=kdata_tmp
        return res
